/* ics_version.h */

#define ICS_VERSION_STR "2022.2.6"
#define ICS_FILEVERSION 2022, 2, 6, 0
#define ICS_NAME "ICSx64.dll"
#define ICS_DESCRIPTION "Image Correlation Spectroscopy Library"
#define ICS_AUTHOR "Christoph Gohlke"
#define ICS_COPYRIGHT "Copyright (C) 2016-2022"
#define ICS_COMPANY \
    "Laboratory for Fluorescence Dynamics, University of California, Irvine"
